# Description
"""
Hive REST API library
Author: HexWay
License: MIT
Copyright 2021, HexWay
"""

# Import
from hive_library import HiveLibrary
from requests import Session, Response, exceptions
from typing import List, Dict, Union, Optional
from uuid import UUID
from dataclasses import dataclass
from datetime import date, timedelta
from json.decoder import JSONDecodeError
from ipaddress import IPv4Address

# Authorship information
__author__ = "HexWay"
__copyright__ = "Copyright 2021, HexWay"
__credits__ = [""]
__license__ = "MIT"
__version__ = "0.0.1b1"
__maintainer__ = "HexWay"
__email__ = "contact@hexway.io"
__status__ = "Development"


# Class AuthenticationError
class AuthenticationError(Exception):
    def __init__(self, message, errors):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.errors = errors


# Class HiveEndpoints
@dataclass
class HiveEndpoints:
    home: str = "/"
    auth: str = "/api/session"
    project: str = "/api/project"
    groups: str = "/api/groups"


# Class HiveRestApi
class HiveRestApi:
    # Set variables

    _endpoints: HiveEndpoints = HiveEndpoints()

    # Init
    def __init__(
        self,
        username: Optional[str] = None,
        password: Optional[str] = None,
        cookie: Optional[str] = None,
        server: Optional[str] = None,
        proxy: Optional[str] = None,
        project_id: Optional[UUID] = None,
        debug: bool = False,
    ) -> None:
        """
        Init HiveRestApi Class
        @param server: Hive server URL string, example: 'http://hive.corp.com:80'
        @param username: Hive username string, example: 'user@mail.com'
        @param password: Hive password string, example: 'Password'
        @param proxy: HTTP Proxy URL string, example: http://127.0.0.1:8080
        @param cookie: Hive cookie string, example: 'SESSIONID=5025db17-a019-4980-9b15-0ffc71a7d0bd'
        @param project_id: Project identifier UUID, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param debug: Run debug mode or not
        """

        # Load variables from config yaml
        config: HiveLibrary.Config = HiveLibrary.load_config()

        # Get server url from config
        if server is None:
            server = config.server

        # Check server is not None
        if server is None:
            print(
                "Hive server URL is not set! Please set server URL, example: 'http://hive.corp.com:80'"
            )
            exit(1)
        else:
            config.server = server

        # Get cookie, username, password from config
        if cookie is None:
            cookie = config.cookie
        if username is None:
            username = config.username
        if password is None:
            password = config.password

        # Save project id to config file
        if project_id is not None:
            config.project_id = project_id

        # Set internal class variables
        self._server = server
        self._debug = debug
        self._session: Session = Session()
        self._session.headers.update(
            {
                "User-Agent": "Hive Client/" + "0.0.1b1",
                "Accept": "application/json",
                "Connection": "close",
            }
        )

        # Set proxy
        if proxy is None:
            proxy = config.proxy
        if proxy is not None:
            self._session.proxies.update(
                {
                    "http": proxy,
                    "https": proxy,
                }
            )
            config.proxy = proxy

        # Authenticate
        try:
            # Check cookie
            if cookie is not None:
                if not self._check_cookie(cookie):
                    if username is not None and password is not None:
                        self._user: Optional[HiveLibrary.User] = self._password_auth(
                            username, password
                        )
                        if self._user is None:
                            raise AuthenticationError(
                                message="Bad creds", errors="Hive REST API auth error"
                            )
                        else:
                            config.username = username
                            config.password = password
                            config.cookie = self._get_cookie()
                    else:
                        raise AuthenticationError(
                            message="Bad cookie", errors="Hive REST API auth error"
                        )
                else:
                    config.cookie = cookie
            # Check username and password
            elif username is not None and password is not None:
                self._user: Optional[HiveLibrary.User] = self._password_auth(
                    username, password
                )
                if self._user is None:
                    raise AuthenticationError(
                        message="Bad username and/or password",
                        errors="Hive REST API auth error",
                    )
                else:
                    config.username = username
                    config.password = password
                    config.cookie = self._get_cookie()
            # Cookie and creds not presented
            else:
                raise AuthenticationError(
                    message="Not found creds or cookies! Please set cookie or username and password.",
                    errors="Hive REST API auth error",
                )

            # Dump variables to config yaml
            HiveLibrary.dump_config(config=config)
        except exceptions.ProxyError as Error:
            print(f"Proxy error: {Error.args[0]}")
        except exceptions.ConnectionError as Error:
            print(f"Connection error: {Error.args[0]}")

    # Internal methods
    @staticmethod
    def _make_error_string(response: Response) -> str:
        return (
            f"\nRequest: {response.request.method} {response.request.url}"
            + f"\nRequest body: {response.request.body}"
            + f"\nResponse status code: {response.status_code}"
            + f"\nResponse headers: {response.headers}"
            + f"\nResponse body: {response.text}"
        )

    def _get_cookie(self) -> str:
        return f'SESSIONID={self._session.cookies.get("SESSIONID")}'

    # Auth
    def _password_auth(
        self, username: str, password: str
    ) -> Optional[HiveLibrary.User]:
        """
        Authorize by username(email) and password on Hive server
        :param username: Hive username string, example: 'user@mail.com'
        :param password: Hive password string, example: 'Password'
        :return: None if error or User object, example:
        HiveLibrary.User(create_date=datetime.datetime(2021, 4, 22, 10, 5, 1, 557699),
                  email='test@mail.com', id=UUID('f04ba8c3-8224-4b3f-b167-b8939ea4f049'),
                  is_admin=True, is_confirmed=True,
                  last_confirmed=datetime.datetime(2021, 4, 22, 10, 5, 1, 925415),
                  login='test@mail.com', name='test@mail.com')
        """
        try:
            response: Response = self._session.post(
                self._server + self._endpoints.auth,
                json={"userEmail": username, "userPassword": password},
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in authorize by user/pass {error_string}"
            assert isinstance(
                response.json(), Dict
            ), f"Bad response in authorize by user/pass {error_string}"
            user_schema: HiveLibrary.User.Schema = HiveLibrary.User.Schema()
            user: HiveLibrary.User = user_schema.load(response.json())
            return user
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def _check_cookie(self, cookie: str) -> bool:
        """
        Authorize cookie
        :param cookie: Hive cookie string, example: 'SESSIONID=5025db17-a019-4980-9b15-0ffc71a7d0bd'
        :return: True if success or False if error
        """
        self._session.cookies.clear()
        self._session.headers.update({"Cookie": cookie})
        response = self._session.get(self._server + self._endpoints.auth)
        return True if response.status_code == 200 else False

    # Get methods
    def get_groups(self) -> Optional[List[HiveLibrary.Group]]:
        """
        Get groups
        :return: None if error or HiveLibrary.Group object, example:
        [HiveLibrary.Group(children=[], create_date=datetime.datetime(2021, 5, 17, 12, 40, 43, 74198),
                           description='default group', full_slug='/default',
                           id=UUID('83f436b8-14fb-40a0-a794-657f600f645a'),
                           last_updated=datetime.datetime(2021, 5, 17, 12, 40, 43, 74198),
                           name='default', parent_id=None, permission_type=None, projects=[], slug='default',
                           users=[HiveLibrary.User(create_date=None, email='test1@test.com',
                                                   id=UUID('aa30af62-649d-43f2-aa51-338b1eac9409'),
                                                   is_admin=False, is_confirmed=False, last_confirmed=None,
                                                   login=None, name='system', permission_type='EDIT', user=None)])]
        """
        try:
            response = self._session.get(
                self._server + self._endpoints.groups + "/list"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in get groups {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in get groups {error_string}"
            group_schema: HiveLibrary.Group.Schema = HiveLibrary.Group.Schema(many=True)
            groups: List[HiveLibrary.Group] = group_schema.load(response.json())
            return groups
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def get_projects_list(self) -> Optional[List[HiveLibrary.Project]]:
        """
        Get projects list
        @return: None if error or Projects list, example:
        [HiveLibrary.Project(projectArchiveDate=None, projectConnectionId=None, projectCreateDate='2021-05-20T16:43:45.895024Z',
                      projectDescription='Unit test project', projectEndDate='2021-01-21T00:00:00.000000Z',
                      projectFullSlug='/default/test_project', projectGroupId='83f436b8-14fb-40a0-a794-657f600f645a',
                      projectId='a4f4f8e5-df18-41ed-9b91-59156ef7f053', projectIsArchived=False,
                      projectName='test_project', projectPermission='ADMIN', projectScope=None,
                      projectSlug='test_project', projectStartDate='2021-01-01T00:00:00.000000Z', projectUsers=None)]
        """
        try:
            response: Response = self._session.get(
                self._server + self._endpoints.project
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in get projects {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in get projects {error_string}"
            project_schema: HiveLibrary.Project.Schema = HiveLibrary.Project.Schema(
                many=True
            )
            projects: List[HiveLibrary.Project] = project_schema.load(response.json())
            return projects
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def get_task(self, project_id: UUID, task_id: UUID) -> Optional[HiveLibrary.Task]:
        """
        Get task information by project id and task id
        @param project_id: Project identifier UUID, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param task_id: Task identifier UUID, example: 'd2ac8146-d9a8-4b38-86b9-b31f20997586'
        @return: None if error or Task object, example:
        HiveLibrary.Task(id=UUID('4ebfb427-f570-4f8c-b6ca-b854c6f46f0d'), type='api_import',
                  user_id=UUID('f04ba8c3-8224-4b3f-b167-b8939ea4f049'),
                  project_id=UUID('05000d7a-d085-48c2-8ffc-24676ee5cdde'),
                  data_source_id=None, file_id=None, file_uuid=None, file_name=None, file_node_id=None,
                  timestamp=datetime.datetime(2021, 5, 24, 8, 48, 26, 705634), state='PROGRESS', total=1, current=0)
        """
        try:
            response = self._session.get(
                self._server
                + self._endpoints.project
                + f"/{project_id}/tasks/{task_id}"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in create project {error_string}"
            assert isinstance(
                response.json(), Dict
            ), f"Bad response in get task {error_string}"
            task_schema: HiveLibrary.Task.Schema = HiveLibrary.Task.Schema()
            task: HiveLibrary.Task = task_schema.load(response.json())
            return task
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def get_task_state(self, project_id: UUID, task_id: UUID) -> Optional[str]:
        """
        Get task state by task id
        :param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        :param task_id: Task identifier string, example: 'd2ac8146-d9a8-4b38-86b9-b31f20997586'
        :return: None if error or task state string, example: 'SUCCESS'
        """
        task: Optional[HiveLibrary.Task] = self.get_task(
            project_id=project_id, task_id=task_id
        )
        if isinstance(task, HiveLibrary.Task):
            return task.state
        else:
            return None

    def task_is_completed(self, project_id: UUID, task_id: UUID) -> bool:
        """
        Check task is completed by task identifier
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param task_id: Task identifier string, example: 'd2ac8146-d9a8-4b38-86b9-b31f20997586'
        @return: True if task is completed or False if task isn't completed or error
        """
        task_state: Optional[str] = self.get_task_state(
            project_id=project_id, task_id=task_id
        )
        if task_state == "SUCCESS":
            return True
        else:
            return False

    def get_ip_labels_for_project(
        self, project_id: UUID
    ) -> Optional[List[HiveLibrary.Label]]:
        """
        Get IP labels for project
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @return: None if error or list of labels, example: [HiveLibrary.Label(count=1, label='192.168.1.1')]
        """
        try:
            response = self._session.get(
                self._server
                + self._endpoints.project
                + f"/{project_id}/graph/nodes/info?label=ip"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in get ip labels {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in get ip labels {error_string}"
            label_schema: HiveLibrary.Label.Schema = HiveLibrary.Label.Schema(many=True)
            labels: List[HiveLibrary.Label] = label_schema.load(response.json())
            return labels
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def get_port_labels_for_project(
        self, project_id: UUID
    ) -> Optional[List[HiveLibrary.Label]]:
        """
        Get port labels for project
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @return: None if error or list of labels, example: [HiveLibrary.Label(count=1, label=80)]
        """
        try:
            response = self._session.get(
                self._server
                + self._endpoints.project
                + f"/{project_id}/graph/nodes/info?label=port"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in get port labels {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in get port labels {error_string}"
            label_schema: HiveLibrary.Label.Schema = HiveLibrary.Label.Schema(many=True)
            labels: List[HiveLibrary.Label] = label_schema.load(response.json())
            return labels
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def get_hosts(self, project_id: UUID) -> Optional[List[HiveLibrary.Host]]:
        """
        Get hosts
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @return: None if error or List of hosts, example:
        [HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])]
        """
        try:
            response = self._session.get(
                self._server + self._endpoints.project + f"/{project_id}/graph"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in get hosts {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in get hosts {error_string}"
            host_schema: HiveLibrary.Project.Schema = HiveLibrary.Host.Schema(many=True)
            hosts: List[HiveLibrary.Host] = host_schema.load(response.json())
            return hosts

        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None

        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def get_file(self, project_id: UUID, file_uuid: UUID) -> Optional[bytes]:
        """
        Get file content from project ID by file uuid
        @param project_id: Project ID string, example: 24a6c90f-dd0b-4b25-bfbe-09784fc3b4ea
        @param file_uuid: File UUID string, example: 66717b2e-9046-4a17-9928-590393a823d4
        @return: Bytes of file content or None if error
        """
        response = self._session.get(
            self._server
            + self._endpoints.project
            + f"/{project_id}/graph/file/{file_uuid}"
        )
        return response.content if response.status_code == 200 else None

    def get_ip_id(
        self, project_id: UUID, ip: IPv4Address = IPv4Address("127.0.0.1")
    ) -> Optional[int]:
        """
        Get node identifier by IP address
        @param project_id: Project ID string, example: '24a6c90f-dd0b-4b25-bfbe-09784fc3b4ea'
        @param ip: IP address, example: '127.0.0.1'
        @return: IP node identifier integer, example: 123 or None if error
        """
        hosts: Optional[List[HiveLibrary.Host]] = self.get_hosts(project_id=project_id)
        if isinstance(hosts, list):
            for host in hosts:
                if host.ip == ip:
                    return host.id
        return None

    def get_port_id(
        self,
        project_id: UUID,
        ip: IPv4Address = IPv4Address("127.0.0.1"),
        port: int = 22,
    ) -> Optional[int]:
        """
        Get node identifier by port number
        @param project_id: Project ID string, example: '24a6c90f-dd0b-4b25-bfbe-09784fc3b4ea'
        @param ip: IP address, example: '127.0.0.1'
        @param port: Port number integer, example: 22
        @return: Port node identifier integer, example: 123 or None if error
        """
        hosts: Optional[List[HiveLibrary.Host]] = self.get_hosts(project_id=project_id)
        if isinstance(hosts, list):
            for host in hosts:
                if host.ip == ip:
                    for host_port in host.ports:
                        if host_port.port == port:
                            return host_port.id
        return None

    def get_project_id_by_name(self, project_name: str) -> Optional[UUID]:
        """
        Get project identifier by name
        @param project_name: Project name string, example: 'project_name'
        @return: Project identifier string or None if error, example: 'be282469-5615-493b-842b-733e6f0b015a'
        """
        projects_list: List[HiveLibrary.Project] = self.get_projects_list()
        for project in projects_list:
            if project.name == project_name and project_name != "":
                return project.id
        return None

    def get_host(self, project_id: UUID, host_id: int) -> Optional[HiveLibrary.Host]:
        """
        Get host by id
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param host_id: Host identifier integer, example: 123
        @return: None if error or Host object, example:
        HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])
        """
        try:
            response = self._session.get(
                self._server
                + self._endpoints.project
                + f"/{project_id}/graph/nodes/{host_id}"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in get host {error_string}"
            assert isinstance(
                response.json(), Dict
            ), f"Bad response in get host {error_string}"
            host_schema: HiveLibrary.Host.Schema = HiveLibrary.Host.Schema()
            host: HiveLibrary.Host = host_schema.load(response.json())
            return host

        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None

        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def get_hostname(
        self, project_id: UUID, hostname_id: int
    ) -> Optional[HiveLibrary.Host.Name]:
        """
        Get hostname by id
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param hostname_id: Hostname identifier integer, example: 123
        @return: None if error or Hostname object, example:
        HiveLibrary.Host.Name(checkmarks=[], files=[], id=124,
                              ips=[HiveLibrary.IP(id=118, ip=IPv4Address('192.168.1.1'), ports=None, tags=None,
                                                  uuid=UUID('edde5d46-a8c5-48c9-9c23-af49e76b0ca1'))],
                              uuid=UUID('21cb06ad-6447-4f25-8ab2-5b21df17a8b8'), notes=None, hostname='unit.test.com',
                              records=[HiveLibrary.Record(children=[],
                                                          create_time=datetime.datetime(2021, 6, 8, 14, 14, 59, 888471),
                                                          creator_uuid=None, extra=None, id=128, uuid=None,
                                                          import_type='unit_test_tool_name',
                                                          name='unit test list hostname record',
                                                          tool_name=None, record_type='list',
                                                          value=['unit test list hostname record value 1',
                                                                 'unit test list hostname record value 2'])],
                              sources=[HiveLibrary.Source(comment=None,
                                                          creator_uuid=UUID('f04ba8c3-8224-4b3f-b167-b8939ea4f049'),
                                                          filename='api_upload', id=114, import_status='Success',
                                                          labels=['Datasource'], name=None,
                                                          post_time=datetime.datetime(2021, 6, 8, 14, 14, 59, 888471),
                                                          type='api', uuid=UUID('1cecbd4c-610b-48ea-9aba-cd6a02537218'))],
                              tags=[HiveLibrary.Tag(id=127, uuid=UUID('8ecae7c0-2024-4332-9ce5-54d539d3e595'),
                                                    name='hostname_tag', parent_id=None, base_node_id=None, labels=[],
                                                    parent_labels=[])])
        """
        try:
            response = self._session.get(
                self._server
                + self._endpoints.project
                + f"/{project_id}/graph/nodes/{hostname_id}"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in get hostname {error_string}"
            assert isinstance(
                response.json(), Dict
            ), f"Bad response in get hostname {error_string}"
            hostname_schema: HiveLibrary.Host.Name.Schema = (
                HiveLibrary.Host.Name.Schema()
            )
            hostname: HiveLibrary.Host.Name = hostname_schema.load(response.json())
            return hostname

        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None

        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def get_port(
        self, project_id: UUID, port_id: int
    ) -> Optional[HiveLibrary.Host.Port]:
        """
        Get port by id
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param port_id: Port identifier integer, example: 123
        @return: None if error or Port object, example:
        HiveLibrary.Host.Port(ip=IPv4Address('192.168.1.1'), checkmarks=[], files=[], id=132,
                              uuid=UUID('3ba90ac2-7c28-4385-b89d-cf3173ed393a'), notes=None, port=12345,
                              service=HiveLibrary.Host.Port.Service(name='http', product='Unit test',
                                                                    version='0.1', cpelist='unit test service cpelist'),
                              protocol='tcp', state='open',
                              records=[HiveLibrary.Record(children=[ .... ],
                                       create_time=datetime.datetime(2021, 6, 8, 14, 43, 59, 271652),
                                       creator_uuid=None, extra=None, id=135, uuid=None,
                                       import_type='unit_test_tool_name', name='unit test nested port record',
                                       tool_name=None, record_type='nested',
                                       value=[ .... ],
                              sources=[HiveLibrary.Source(comment=None,
                                                          creator_uuid=UUID('f04ba8c3-8224-4b3f-b167-b8939ea4f049'),
                                                          filename='api_upload', id=114, import_status='Success',
                                                          labels=['Datasource'], name=None,
                                                          post_time=datetime.datetime(2021, 6, 8, 14, 43, 59, 271652),
                                                          type='api', uuid=UUID('6b08ca3e-706f-4ccf-915a-daced88e97d5'))],
                              tags=[HiveLibrary.Tag(id=134, uuid=None, name='port_tag', parent_id=None,
                                    base_node_id=None, labels=[], parent_labels=[])])
        """
        try:
            response = self._session.get(
                self._server
                + self._endpoints.project
                + f"/{project_id}/graph/nodes/{port_id}"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in get port {error_string}"
            assert isinstance(
                response.json(), Dict
            ), f"Bad response in get port {error_string}"
            port_schema: HiveLibrary.Host.Port.Schema = HiveLibrary.Host.Port.Schema()
            port: HiveLibrary.Host.Port = port_schema.load(response.json())
            return port

        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None

        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    # Search methods
    def search(
        self, project_id: UUID, search_string: str
    ) -> Optional[List[HiveLibrary.Host]]:
        """
        Search hosts in project
        :param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        :param search_string: Search string, example: 'ip == 127.0.0.1 and port == 22'
        :return: None if error or Hosts list, example:
        [HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])]
        """
        try:
            response = self._session.post(
                self._server + self._endpoints.project + f"/{project_id}/graph/search",
                json={"searchString": search_string},
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in search: {search_string} {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in search: {search_string} {error_string}"
            host_schema: HiveLibrary.Project.Schema = HiveLibrary.Host.Schema(many=True)
            hosts: List[HiveLibrary.Host] = host_schema.load(response.json())
            return hosts
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def search_by_ip(
        self, project_id: UUID, ip: Union[IPv4Address, List[IPv4Address]]
    ) -> Optional[List[HiveLibrary.Host]]:
        """
        Get all hosts by IP address or IP addresses list
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param ip: IP address string or IP addresses list, example: '1.1.1.1' or ['1.1.1.1', '8.8.8.8']
        @return: None if error or Hosts list, example:
        [HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])]
        """
        if isinstance(ip, List):
            for i in range(len(ip)):
                ip[i] = f"ip == '{ip[i]}'"
            search: str = " or ".join(ip)
        else:
            search: str = f"ip == '{ip}'"
        return self.search(project_id, search)

    def search_by_port(
        self, project_id: UUID, port: Union[int, List[int]]
    ) -> Optional[List[HiveLibrary.Host]]:
        """
        Get all hosts by Port or Ports list
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param port: Port or ports list, example: '22' or ['22', '80']
        @return: None if error or Hosts list, example:
        [HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])]
        """
        if isinstance(port, List):
            for i in range(len(port)):
                port[i] = f"port == {port[i]}"
            search: str = " or ".join(port)
        else:
            search: str = f"port == {port}"
        return self.search(project_id, search)

    def search_by_ip_and_port(
        self, project_id: UUID, ip: IPv4Address, port: int
    ) -> Optional[List[HiveLibrary.Host]]:
        """
        Get all hosts by IP address and Port
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param port: IP address string, example: '127.0.0.1'
        @param ip: Port number string, example: '22'
        @return: None if error or Hosts list, example:
        [HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])]
        """
        search: str = f"ip == {ip} and port == {port}"
        return self.search(project_id, search)

    def search_by_hostname(
        self, project_id: UUID, hostname: Union[str, List[str]]
    ) -> Optional[List[HiveLibrary.Host]]:
        """
        Get all hosts by Hostname or hostnames list
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param hostname: Hostname or hostnames list, example: 'test.host.com' or ['test1.host.com', 'test2.host.com']
        @return: None if error or Hosts list, example:
        [HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])]
        """
        if isinstance(hostname, List):
            for i in range(len(hostname)):
                hostname[i] = f"hostname == '{hostname[i]}'"
            search: str = " or ".join(hostname)
        else:
            search: str = f"hostname == '{hostname}'"
        return self.search(project_id, search)

    def search_by_tag(
        self, project_id: UUID, tag: Union[str, List[str]]
    ) -> Optional[List[HiveLibrary.Host]]:
        """
        Get all hosts by Tag or tags list
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param tag: Tag or tags list, example: 'tag' or ['tag1', 'tag2']
        @return: None if error or Hosts list, example:
        [HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])]
        """
        if isinstance(tag, List):
            for i in range(len(tag)):
                tag[i] = f"tag == '{tag[i]}'"
            search: str = " or ".join(tag)
        else:
            search: str = f"tag == '{tag}'"
        return self.search(project_id, search)

    def search_by_service(
        self, project_id: UUID, service: Union[str, List[str]]
    ) -> Optional[List[HiveLibrary.Host]]:
        """
        Get all hosts by Tag or service or services list
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param service: Tag or tags list, example: 'ssh' or ['ssh', 'http']
        @return: None if error or Hosts list, example:
        [HiveLibrary.Host(checkmarks=[], files=[], id=106, uuid=UUID('cc862d8d-39ad-49fa-918f-3762847ecdf2'),
                   notes=None, ip=IPv4Address('192.168.1.1'), records=[],
                   hostnames=[HiveLibrary.Host.Name(checkmarks=[], files=[], id=182,
                              uuid=UUID('459240f7-9825-45ca-af4b-86bcb4ac5b0c'), notes=None,
                              hostname='unit.test.com', records=[],
                              tags=[HiveLibrary.Tag(id=186, uuid=None, name='hostname_tag')])],
                   ports=[HiveLibrary.Host.Port(checkmarks=[], files=[], id=190,
                          uuid=UUID('4ad7c296-995b-4890-a470-5191708b087b'), notes=None, port=1234,
                          service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                         product='unit test service name',
                                                         version='0.1',
                                                         cpelist='unit test service cpelist'),
                          protocol='tcp', state='open', records=[],
                          tags=[HiveLibrary.Tag(id=192, uuid=UUID('03db99e5-3fa4-452d-95d9-f3982b9c641f'), name='port_tag')])],
                   tags=[HiveLibrary.Tag(id=188, uuid=None, name='host_tag')])]
        """
        if isinstance(service, List):
            for i in range(len(service)):
                service[i] = f"service == '{service[i]}'"
            search: str = " or ".join(service)
        else:
            search: str = f"service == '{service}'"
        return self.search(project_id, search)

    # Post methods
    def create_project(
        self,
        project: HiveLibrary.Project,
    ) -> Optional[HiveLibrary.Project]:
        """
        Create new project
        :param project: HiveLibrary.Project object, example: HiveLibrary.Project(projectArchiveDate=None,
                                                                   projectConnectionId=None,
                                                                   projectCreateDate=None,
                                                                   projectDescription='Unit test project',
                                                                   projectEndDate='2021-01-21',
                                                                   projectFullSlug=None,
                                                                   projectGroupId='83f436b8-14fb-40a0-a794-657f600f645a',
                                                                   projectId=None,
                                                                   projectIsArchived=False,
                                                                   projectName='test_project',
                                                                   projectPermission=None,
                                                                   projectScope=None,
                                                                   projectSlug=None,
                                                                   projectStartDate='2021-01-01',
                                                                   projectUsers=None)
        :return: None if error or Project object, example:
        HiveLibrary.Project(projectArchiveDate=None, projectConnectionId=None, projectCreateDate='2021-05-20T16:30:03.120592Z',
                     projectDescription='Unit test project', projectEndDate='2021-01-21T00:00:00.000000Z',
                     projectFullSlug='/default/test_project', projectGroupId='83f436b8-14fb-40a0-a794-657f600f645a',
                     projectId='78171434-ceac-4bca-a1ba-a594a7c795d9', projectIsArchived=False,
                     projectName='test_project', projectPermission='OWNER', projectScope=None,
                     projectSlug='test_project', projectStartDate='2021-01-01T00:00:00.000000Z',
                     projectUsers=[HiveLibrary.Project.User(permissionType='OWNER',
                                                     userEmail='user@mail.com',
                                                     userName='user@mail.com',
                                                     userId='f04ba8c3-8224-4b3f-b167-b8939ea4f049')])
        """
        try:
            if project.start_date is None:
                project.start_date = date.today()
            if project.end_date is None:
                project.end_date = date.today() + timedelta(weeks=1)
            assert (
                project.name is not None
            ), "For create project, please set project Name"
            if project.group_id is None:
                groups: Optional[List[HiveLibrary.Group]] = self.get_groups()
                assert groups is not None, "Failed to get groups for create project"
                project.group_id = groups[0].id
            response: Response = self._session.post(
                self._server + self._endpoints.project,
                json=HiveLibrary.Project.Schema().dump(project),
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in create project {error_string}"
            assert isinstance(
                response.json(), Dict
            ), f"Bad response in create project {error_string}"
            project_schema: HiveLibrary.Project.Schema = HiveLibrary.Project.Schema()
            project: HiveLibrary.Project = project_schema.load(response.json())
            return project
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def import_from_file(
        self,
        file_location: str,
        import_type: str = "nmap",
        project_id: Optional[UUID] = None,
        project_name: Optional[str] = None,
    ) -> Optional[UUID]:
        """
        Import data from file
        :param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        :param project_name: Project name string, example: 'test_project'
        :param import_type: Import type string, examples: 'nmap', 'nessus', 'metasploit', 'cobaltstrike'
        :param file_location: File location string, example: '/tmp/test.xml'
        :return: None if error or import task identifier, example: 'e08d16f3-b864-44ac-a90c-8e10f4af9893'
        """
        try:
            if project_name is None and project_id is None:
                assert False, "Project ID and Name is not set"
            if project_name is not None:
                project_id = self.get_project_id_by_name(project_name=project_name)
            with open(file_location, "r") as file:
                file_content = file.read()
            response = self._session.post(
                self._server
                + self._endpoints.project
                + f"/{project_id}/graph?importType={import_type}",
                data=file_content,
                headers={"Content-Type": "applications/octet-stream"},
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in import from file {error_string}"
            if "taskId" in response.json():
                if isinstance(response.json()["taskId"], str):
                    return UUID(response.json()["taskId"])
            return None
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def create_hosts(
        self, project_id: UUID, hosts: List[HiveLibrary.Host]
    ) -> Optional[UUID]:
        """
        Create hosts in project
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param hosts: List of hosts, example:
        [HiveLibrary.Host(ipv4='192.168.1.1',
                   records=[HiveLibrary.Host.Record(name='unit test nested host record',
                                             tool_name='unit_test_tool_name',
                                             record_type='nested',
                                             value=[HiveLibrary.Host.Record(name='unit test host record name 1',
                                                                     tool_name='unit_test_tool_name',
                                                                     record_type='string',
                                                                     value='unit test host record value 1'),
                                                   HiveLibrary.Host.Record(name='unit test host record name 2',
                                                                    tool_name='unit_test_tool_name',
                                                                    record_type='string',
                                                                    value='unit test host record value 2')])],
                   hostnames=[HiveLibrary.Host.Hostname(hostname='unit.test.com',
                                                 records=[HiveLibrary.Host.Record(name='unit test list hostname record',
                                                                           tool_name='unit_test_tool_name',
                                                                           record_type='list',
                                                                           value=['unit test list hostname record 1',
                                                                                  'unit test list hostname record 2'])])],
                   ports=[HiveLibrary.Host.Port(port=1234,
                                         service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                                        product='unit test service name',
                                                                        version='0.1'),
                                         protocol='tcp',
                                         state='open',
                                         records=[HiveLibrary.Host.Record(name='unit test string port record',
                                                                   tool_name='unit_test_tool_name',
                                                                   record_type='string',
                                                                   value='unit test string port record value')])])]
        @return: None if error or import task UUID
        """
        try:
            hosts_schema: HiveLibrary.Host.Schema = HiveLibrary.Host.Schema(many=True)
            response = self._session.post(
                self._server + self._endpoints.project + f"/{project_id}/graph/api",
                json=hosts_schema.dump(hosts),
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in create hosts {error_string}"
            assert isinstance(
                response.json(), Dict
            ), f"Bad response in create hosts {error_string}"
            assert (
                "taskId" in response.json()
            ), f"Not found key taskId in create hosts response {error_string}"
            assert isinstance(
                response.json()["taskId"], str
            ), f"Bad value for key taskId in create hosts response {error_string}"
            return UUID(response.json()["taskId"])
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def create_host(self, project_id: UUID, host: HiveLibrary.Host) -> Optional[UUID]:
        """
        Create host in project
        @param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        @param host: Host object, example:
        HiveLibrary.Host(ipv4='192.168.1.1',
                  records=[HiveLibrary.Host.Record(name='unit test nested host record',
                                            tool_name='unit_test_tool_name',
                                            record_type='nested',
                                            value=[HiveLibrary.Host.Record(name='unit test host record name 1',
                                                                    tool_name='unit_test_tool_name',
                                                                    record_type='string',
                                                                    value='unit test host record value 1'),
                                                  HiveLibrary.Host.Record(name='unit test host record name 2',
                                                                   tool_name='unit_test_tool_name',
                                                                   record_type='string',
                                                                   value='unit test host record value 2')])],
                  hostnames=[HiveLibrary.Host.Hostname(hostname='unit.test.com',
                                                records=[HiveLibrary.Host.Record(name='unit test list hostname record',
                                                                          tool_name='unit_test_tool_name',
                                                                          record_type='list',
                                                                          value=['unit test list hostname record 1',
                                                                                 'unit test list hostname record 2'])])],
                  ports=[HiveLibrary.Host.Port(port=1234,
                                        service=HiveLibrary.Host.Port.Service(name='unit test service name',
                                                                       product='unit test service name',
                                                                       version='0.1'),
                                        protocol='tcp',
                                        state='open',
                                        records=[HiveLibrary.Host.Record(name='unit test string port record',
                                                                  tool_name='unit_test_tool_name',
                                                                  record_type='string',
                                                                  value='unit test string port record value')])])
        @return: None if error or import task UUID
        """
        return self.create_hosts(project_id=project_id, hosts=[host])

    def create_note(
        self, note_text: str, project_id: UUID, node_id: int
    ) -> Optional[HiveLibrary.Note]:
        """
        Create note for node in project
        :param note_text: Note text string, example: 'test'
        :param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        :param node_id: Project identifier integer, example: 123
        :return: None if error or Note object, example:
        HiveLibrary.Note(base_node_id=111, id=129, parent_id=111, type='Node',
                  uuid=UUID('1da9570d-00d7-4044-8d4b-46cdbd7d57b1'),
                  create_time=datetime.datetime(2021, 5, 31, 13, 24, 6, 560968),
                  creator_uuid=UUID('f04ba8c3-8224-4b3f-b167-b8939ea4f049'),
                  files=[], text='unit test note text', node_id=None,
                  checkmarks=[], labels=['Note'], tags=[])
        """
        try:
            response: Response = self._session.post(
                self._server + self._endpoints.project + f"/{project_id}/graph/nodes",
                json=[{"nodeId": node_id, "text": note_text}],
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in create note {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in create note {error_string}"
            assert (
                len(response.json()) == 1
            ), f"Bad response in create note {error_string}"
            note_schema: HiveLibrary.Note.Schema = HiveLibrary.Note.Schema(many=True)
            note: HiveLibrary.Note = note_schema.load(response.json())[0]
            return note
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def create_tag(
        self, tag_name: str, project_id: UUID, node_id: int
    ) -> Optional[HiveLibrary.Tag]:
        """
        Create tag for node in project
        :param tag_name: Tag name string, example: 'test_tag'
        :param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        :param node_id: Project identifier integer, example: 123
        :return: None if error or Tag object, example:
        HiveLibrary.Tag(id=165, uuid=None, name='unit_test_tag_name',
                 parent_id=145, base_node_id=145,
                 labels=['Tag'], parent_labels=['Ip'])
        """
        try:
            response: Response = self._session.post(
                self._server + self._endpoints.project + f"/{project_id}/graph/nodes",
                json=[{"nodeId": node_id, "tag": tag_name}],
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in create tag {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in create tag {error_string}"
            assert (
                len(response.json()) == 1
            ), f"Bad response in create tag {error_string}"
            tag_schema: HiveLibrary.Tag.Schema = HiveLibrary.Tag.Schema()
            tag: HiveLibrary.Tag = tag_schema.load(response.json()[0])
            return tag
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def upload_file(
        self,
        file_name: str,
        file_content: bytes,
        project_id: UUID,
        node_id: int,
        file_caption: Optional[str] = None,
    ) -> Optional[HiveLibrary.File]:
        """
        Upload file for node in project
        :param file_name: File name string, example: 'test.txt'
        :param file_content: Bytes of file content, example: b'Hello World!'
        :param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        :param node_id: Project identifier integer, example: 123
        :return: None if error or File object, example:
        HiveLibrary.File(base_node_id=None, caption=None, id=133, type=None,
                  uuid=UUID('b7a2eed4-661f-4797-ad3b-9015bf3e73b7'),
                  post_time=datetime.datetime(2021, 5, 31, 13, 30, 57, 944848),
                  create_time=None, user_uuid=UUID('f04ba8c3-8224-4b3f-b167-b8939ea4f049'), creator_uuid=None,
                  control_sum='sha256:747153e0f6b14eb609fd7cb6921fa871b08f26fec6e042075a2ab30a1af4d295',
                  name='unit_test_file.txt', size=22, mime_type='text/plain', node_id=None, parent_id=111,
                  labels=['File'], parent_labels=['Ip'])
        """
        try:
            response: Response = self._session.post(
                self._server + self._endpoints.project + f"/{project_id}/graph/file",
                data=file_content,
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in upload file {error_string}"
            assert isinstance(
                response.json(), Dict
            ), f"Bad response in upload file {error_string}"
            file_schema: HiveLibrary.File.Schema = HiveLibrary.File.Schema()
            file: HiveLibrary.File = file_schema.load(response.json())
            file.name = file_name
            file.node_id = node_id
            file.caption = file_caption
            response: Response = self._session.post(
                self._server + self._endpoints.project + f"/{project_id}/graph/nodes",
                json=[file_schema.dump(file)],
            )
            assert (
                response.status_code == 200
            ), f"Bad status code in upload file {error_string}"
            assert isinstance(
                response.json(), List
            ), f"Bad response in upload file {error_string}"
            assert (
                len(response.json()) == 1
            ), f"Bad response in create note {error_string}"
            file_schema: HiveLibrary.File.Schema = HiveLibrary.File.Schema()
            file: HiveLibrary.File = file_schema.load(response.json()[0])
            return file
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    # Delete methods
    def delete_project_by_id(self, project_id: UUID) -> Optional[HiveLibrary.Project]:
        """
        Delete project by identifier
        :param project_id: Project identifier string, example: 'be282469-5615-493b-842b-733e6f0b015a'
        :return: None if error or Project object, example:
        HiveLibrary.Project(projectArchiveDate=None, projectConnectionId=None, projectCreateDate='2021-05-20T16:30:03.120592Z',
                     projectDescription='Unit test project', projectEndDate='2021-01-21T00:00:00.000000Z',
                     projectFullSlug='/default/test_project', projectGroupId='83f436b8-14fb-40a0-a794-657f600f645a',
                     projectId='78171434-ceac-4bca-a1ba-a594a7c795d9', projectIsArchived=False,
                     projectName='test_project', projectPermission='OWNER', projectScope=None,
                     projectSlug='test_project', projectStartDate='2021-01-01T00:00:00.000000Z',
                     projectUsers=[HiveLibrary.Project.User(permissionType='OWNER',
                                                     userEmail='user@mail.com',
                                                     userName='user@mail.com',
                                                     userId='f04ba8c3-8224-4b3f-b167-b8939ea4f049')])
        """
        try:
            response = self._session.delete(
                self._server + self._endpoints.project + f"/{project_id}"
            )
            error_string: str = ""
            if self._debug:
                error_string = self._make_error_string(response)
            assert (
                response.status_code == 200
            ), f"Bad status code in delete project {error_string}"
            project_schema: HiveLibrary.Project.Schema = HiveLibrary.Project.Schema()
            project: HiveLibrary.Project = project_schema.load(response.json())
            return project
        except AssertionError as error:
            print(f"Assertion error: {error.args[0]}")
            return None
        except JSONDecodeError as error:
            print(f"JSON Decode error: {error.args[0]}")
            return None

    def delete_project_by_name(self, project_name: str):
        """
        Delete project by name
        :param project_name: Project name string, example: 'msf_project'
        :return: None if error or Project object, example:
        HiveLibrary.Project(projectArchiveDate=None, projectConnectionId=None, projectCreateDate='2021-05-20T16:30:03.120592Z',
                     projectDescription='Unit test project', projectEndDate='2021-01-21T00:00:00.000000Z',
                     projectFullSlug='/default/test_project', projectGroupId='83f436b8-14fb-40a0-a794-657f600f645a',
                     projectId='78171434-ceac-4bca-a1ba-a594a7c795d9', projectIsArchived=False,
                     projectName='test_project', projectPermission='OWNER', projectScope=None,
                     projectSlug='test_project', projectStartDate='2021-01-01T00:00:00.000000Z',
                     projectUsers=[HiveLibrary.Project.User(permissionType='OWNER',
                                                     userEmail='user@mail.com',
                                                     userName='user@mail.com',
                                                     userId='f04ba8c3-8224-4b3f-b167-b8939ea4f049')])
        """
        project_id: Optional[UUID] = None
        projects_list = self.get_projects_list()
        for project in projects_list:
            if project.name == project_name:
                project_id = project.id
        if project_id is None:
            return None
        else:
            return self.delete_project_by_id(project_id=project_id)



# Description
"""
Hive dataclasses
Author: HexWay
License: MIT
Copyright 2021, HexWay
"""

# Import
from hive_library.enum import RecordTypes, PermissionTypes, TaskStates
from typing import List, Dict, Optional, Union
from uuid import UUID
from dataclasses import dataclass, field
from marshmallow import (
    fields,
    pre_load,
    post_load,
    pre_dump,
    post_dump,
    validate,
    EXCLUDE,
)
from marshmallow import Schema as MarshmallowSchema
from datetime import datetime
from ipaddress import ip_address, IPv4Address
from yaml import safe_load, safe_dump
from os import path
from os import makedirs
from pathlib import Path


# Authorship information
__author__ = "HexWay"
__copyright__ = "Copyright 2021, HexWay"
__credits__ = [""]
__license__ = "MIT"
__version__ = "0.0.1b1"
__maintainer__ = "HexWay"
__email__ = "contact@hexway.io"
__status__ = "Development"


PermissionTypes = PermissionTypes.list()
RecordTypes = RecordTypes.list()
TaskStates = TaskStates.list()


class HiveLibrary:
    @dataclass
    class Config:
        file: str = f"{str(Path.home())}/.hive/config.yaml"
        server: Optional[str] = None
        proxy: Optional[str] = None
        username: Optional[str] = None
        password: Optional[str] = None
        cookie: Optional[str] = None
        project_id: Optional[UUID] = None

        class Schema(MarshmallowSchema):
            server = fields.String(missing=None)
            proxy = fields.String(missing=None)
            username = fields.String(missing=None)
            password = fields.String(missing=None)
            cookie = fields.String(missing=None)
            project_id = fields.UUID(missing=None)

            @post_dump(pass_many=False)
            def clean_missing(self, data, many, **kwargs):
                clean_data = data.copy()
                for key in filter(lambda key: data[key] is None, data):
                    del clean_data[key]
                return clean_data

            @post_load
            def make_config(self, data, **kwargs):
                return HiveLibrary.Config(**data)

    @dataclass
    class User:
        create_date: Optional[datetime] = None
        email: Optional[str] = None
        id: Optional[UUID] = None
        is_admin: bool = False
        is_confirmed: bool = False
        last_confirmed: Optional[datetime] = None
        login: Optional[str] = None
        name: Optional[str] = None
        permission_type: Optional[str] = None
        user: Optional[Dict[str, str]] = None

        class Schema(MarshmallowSchema):
            permission_type = fields.String(
                validate=validate.OneOf(PermissionTypes),
                data_key="permissionType",
                missing=None,
            )
            user = fields.Dict(missing=None)
            create_date = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
                missing=None,
                data_key="userCreateDate",
            )
            email = fields.String(data_key="userEmail")
            id = fields.UUID(required=True, data_key="userId")
            is_admin = fields.Bool(default=False, data_key="userIsAdmin")
            is_confirmed = fields.Bool(default=False, data_key="userIsConfirmed")
            last_confirmed = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
                missing=None,
                data_key="userLastConfirmed",
            )
            login = fields.Str(missing=None, data_key="userLogin")
            name = fields.Str(missing=None, data_key="userName")

            @post_load
            def make_user(self, data, **kwargs):
                return HiveLibrary.User(**data)

    @dataclass
    class Group:
        def __str__(self):
            return self.name

        children: Optional[List["HiveLibrary.Group"]] = field(
            default_factory=lambda: []
        )
        create_date: Optional[datetime] = None
        description: Optional[str] = None
        full_slug: Optional[str] = None
        id: Optional[UUID] = None
        last_updated: Optional[datetime] = None
        name: Optional[str] = None
        parent_id: Optional[str] = None
        permission_type: Optional[str] = None
        projects: Optional[List["HiveLibrary.Project"]] = field(
            default_factory=lambda: []
        )
        slug: Optional[str] = None
        users: Optional[List["HiveLibrary.User"]] = field(default_factory=lambda: [])

        class Schema(MarshmallowSchema):
            permission_type = fields.String(
                validate=validate.OneOf(PermissionTypes),
                missing=None,
                data_key="permissionType",
            )
            parent_id = fields.UUID(allow_none=True, missing=None, data_key="parentId")
            id = fields.UUID(required=True)
            description = fields.String(missing=None)
            name = fields.String(required=True)
            create_date = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ", required=True, data_key="createDate"
            )
            last_updated = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ", required=True, data_key="lastUpdated"
            )
            slug = fields.String(missing=None)
            full_slug = fields.String(missing=None, data_key="fullSlug")
            projects = fields.Nested(lambda: HiveLibrary.Project.Schema, many=True)
            children = fields.Nested(lambda: HiveLibrary.Group.Schema, many=True)
            users = fields.Nested(lambda: HiveLibrary.User.Schema, many=True)

            @post_load
            def make_group(self, data, **kwargs):
                return HiveLibrary.Group(**data)

    @dataclass
    class Project:
        permission: Optional[str] = None
        group_id: Optional[UUID] = None
        id: Optional[UUID] = None
        description: Optional[str] = None
        name: Optional[str] = None
        create_date: Optional[datetime] = None
        is_archived: bool = False
        start_date: Optional[datetime] = None
        end_date: Optional[datetime] = None
        archive_date: Optional[datetime] = None
        hawser_id: Optional[UUID] = None
        scope: Optional[str] = None
        slug: Optional[str] = None
        full_slug: Optional[str] = None
        users: Optional[List["HiveLibrary.User"]] = field(default_factory=lambda: [])

        def __str__(self):
            return self.name

        class Schema(MarshmallowSchema):
            permission = fields.String(
                validate=validate.OneOf(PermissionTypes),
                load_only=True,
                missing=None,
                data_key="projectPermission",
                nullable=True,
            )
            group_id = fields.UUID(
                allow_none=True, missing=None, data_key="projectGroupId"
            )
            id = fields.UUID(load_only=True, required=True, data_key="projectId")
            description = fields.String(missing=None, data_key="projectDescription")
            name = fields.String(required=True, data_key="projectName")
            create_date = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
                required=True,
                load_only=True,
                data_key="projectCreateDate",
            )
            is_archived = fields.Boolean(load_only=True, data_key="projectIsArchived")
            start_date = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ", data_key="projectStartDate"
            )
            end_date = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ", data_key="projectEndDate"
            )
            archive_date = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
                load_only=True,
                missing=None,
                data_key="projectArchiveDate",
            )
            hawser_id = fields.UUID(
                load_only=True,
                allow_none=True,
                missing=None,
                data_key="projectConnectionId",
            )
            scope = fields.String(load_only=True, missing=None, data_key="projectScope")
            slug = fields.String(load_only=True, missing=None, data_key="projectSlug")
            full_slug = fields.String(
                load_only=True, missing=None, data_key="projectFullSlug"
            )
            users = fields.Nested(
                lambda: HiveLibrary.User.Schema,
                load_only=True,
                missing=None,
                many=True,
                data_key="projectUsers",
            )

            @post_dump(pass_many=False)
            def clean_missing(self, data, many, **kwargs):
                clean_data = data.copy()
                for key in filter(lambda key: data[key] is None, data):
                    del clean_data[key]
                return clean_data

            @post_load
            def make_project(self, data, **kwargs):
                return HiveLibrary.Project(**data)

    @dataclass
    class Task:
        id: Optional[UUID] = None
        type: Optional[str] = None
        user_id: Optional[UUID] = None
        project_id: Optional[UUID] = None
        data_source_id: Optional[int] = None
        file_id: Optional[int] = None
        file_uuid: Optional[UUID] = None
        file_name: Optional[str] = None
        file_node_id: Optional[int] = None
        timestamp: Optional[datetime] = None
        state: Optional[str] = None
        total: Optional[int] = None
        current: Optional[int] = None
        exc_message: Optional[str] = None
        exc_type: Optional[str] = None

        class Schema(MarshmallowSchema):
            id = fields.UUID(required=True, data_key="taskId")
            type = fields.String(required=True, data_key="taskType")
            user_id = fields.UUID(required=True, data_key="userId")
            project_id = fields.UUID(required=True, data_key="projectId")
            data_source_id = fields.Integer(missing=None, data_key="datasourceId")
            file_id = fields.Integer(missing=None, data_key="fileId")
            file_uuid = fields.UUID(missing=None, data_key="fileUuid")
            file_name = fields.String(missing=None, data_key="filename")
            file_node_id = fields.Integer(missing=None, data_key="filenodeId")
            timestamp = fields.DateTime("%Y-%m-%dT%H:%M:%S.%fZ", required=True)
            state = fields.String(
                validate=validate.OneOf(TaskStates),
                allow_none=False,
                required=True,
            )
            total = fields.Integer(required=False, default=None, missing=None)
            current = fields.Integer(required=False, default=None, missing=None)
            exc_message = fields.String(missing=None)
            exc_type = fields.String(missing=None)

            @post_load
            def make_task(self, data, **kwargs):
                return HiveLibrary.Task(**data)

    @dataclass
    class Label:
        count: Optional[int] = None
        label: Union[None, int, str, IPv4Address] = None

        class Schema(MarshmallowSchema):
            count = fields.Integer(default=None, allow_none=True)
            label = fields.Raw(default=None, allow_none=True)

            @post_load
            def make_label(self, data, **kwargs):
                try:
                    if isinstance(data["label"], str):
                        data["label"] = ip_address(data["label"])
                except ValueError:
                    pass
                return HiveLibrary.Label(**data)

    @dataclass
    class Tag:
        parent_id: Optional[int] = None
        id: Optional[int] = None
        name: Optional[str] = None

        class Schema(MarshmallowSchema):
            parent_id = fields.Integer(missing=None, default=None, data_key="parentId")
            id = fields.Integer(default=None, allow_none=True)
            name = fields.String(default=None, allow_none=True)

            class Meta:
                unknown = EXCLUDE

            @post_dump(pass_many=True)
            def parse_tags(self, data, many, **kwargs):
                if len(data) > 0:
                    tags: List[str] = list()
                    for tag in data:
                        tags.append(tag["name"])
                    return tags
                else:
                    return data

            @post_load
            def make_tag(self, data, **kwargs):
                return HiveLibrary.Tag(**data)

    @dataclass
    class Checkmark:
        id: Optional[int] = None
        uuid: Optional[UUID] = None
        type: Optional[str] = None
        create_time: Optional[datetime] = None
        creator_uuid: Optional[UUID] = None
        editor_uuid: Optional[UUID] = None
        description: Optional[str] = None
        done: bool = False
        synchronizable: bool = False
        name: Optional[str] = None
        order: Optional[int] = None
        children: Optional[List["HiveLibrary.Checkmark"]] = field(
            default_factory=lambda: []
        )

        class Schema(MarshmallowSchema):
            id = fields.Integer(
                load_only=True, missing=None, default=None, data_key="_id"
            )
            uuid = fields.UUID(default=None, allow_none=True)
            type = fields.String(
                load_only=True, missing=None, default=None, data_key="_type"
            )
            create_time = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
                load_only=True,
                data_key="createTime",
                missing=None,
            )
            creator_uuid = fields.UUID(
                load_only=True, missing=None, default=None, data_key="creatorUuid"
            )
            editor_uuid = fields.UUID(
                load_only=True, missing=None, default=None, data_key="editorUuid"
            )
            description = fields.String(default=None, allow_none=True)
            done = fields.Bool(default=False)
            synchronizable = fields.Bool(default=False)
            name = fields.String(default=None, allow_none=True)
            order = fields.Integer(
                load_only=True, missing=None, default=None, data_key="order"
            )
            children = fields.Nested(
                lambda: HiveLibrary.Checkmark.Schema,
                many=True,
                missing=[],
                default=[],
                allow_none=True,
                data_key="childcheckmark",
            )

            class Meta:
                unknown = EXCLUDE

            @pre_load(pass_many=False)
            def pre_make_checkmark(self, data, **kwargs):
                if "id" in data:
                    data["_id"] = data["id"]
                    del data["id"]
                return data

            @post_load
            def make_checkmark(self, data, **kwargs):
                return HiveLibrary.Checkmark(**data)

    @dataclass
    class Record:
        children: Optional[List["HiveLibrary.Record"]] = field(
            default_factory=lambda: []
        )
        create_time: Optional[datetime] = None
        creator_uuid: Optional[UUID] = None
        extra: Optional[str] = None
        id: Optional[int] = None
        uuid: Optional[UUID] = None
        import_type: Optional[str] = None
        name: Optional[str] = None
        tool_name: Optional[str] = None
        record_type: Optional[str] = None
        value: Union[
            None, int, str, List[Union[int, str, "HiveLibrary.Record"]]
        ] = field(default_factory=lambda: [])

        class Schema(MarshmallowSchema):
            children = fields.Nested(
                lambda: HiveLibrary.Record.Schema,
                load_only=True,
                many=True,
                missing=[],
                default=[],
            )
            create_time = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
                load_only=True,
                data_key="createTime",
                missing=None,
            )
            creator_uuid = fields.UUID(
                data_key="creatorUuid", load_only=True, missing=None
            )
            extra = fields.String(load_only=True, missing=None)
            id = fields.Integer(load_only=True, missing=None)
            uuid = fields.UUID(load_only=True, missing=None)
            import_type = fields.String(
                data_key="importType", load_only=True, missing=None
            )
            name = fields.String(missing=None)
            tool_name = fields.String(missing=None, default="nmap")
            record_type = fields.String(
                validate=validate.OneOf(RecordTypes),
            )
            value = fields.Raw(missing=None)

            @pre_dump(pass_many=True)
            def parse_records(self, data, many, **kwargs):
                if isinstance(data, List):
                    result_list: List[Dict] = list()
                    for record in data:
                        if isinstance(record, HiveLibrary.Record):
                            result: Dict = dict()
                            for key in ["name", "tool_name", "record_type"]:
                                result[key] = record.__dict__[key]
                            result["value"] = self.parse_records(
                                record.__dict__["value"], True
                            )
                            result_list.append(result)
                        else:
                            result_list.append(record)
                    return result_list
                else:
                    if isinstance(data, HiveLibrary.Record):
                        result: Dict = dict()
                        for key in ["name", "tool_name", "record_type"]:
                            result[key] = data.__dict__[key]
                        result["value"] = self.parse_records(
                            data.__dict__["value"], True
                        )
                        return result
                    else:
                        return data

            @pre_load(pass_many=False)
            def pre_make_record(self, data, **kwargs):
                for key in data:
                    if key == "recordType":
                        data["record_type"] = data["recordType"]
                        del data["recordType"]
                        break
                return data

            @post_load
            def make_record(self, data, **kwargs):
                if isinstance(data["value"], List):
                    new_values: List = list()
                    for value in data["value"]:
                        if isinstance(value, Dict):
                            new_values.append(self.make_record(value))
                        else:
                            new_values.append(value)
                    data["value"] = new_values
                if "children" in data:
                    if len(data["children"]) > 0 and data["value"] is None:
                        data["value"] = data["children"]
                return HiveLibrary.Record(**data)

    @dataclass
    class Source:
        id: Optional[int] = None
        uuid: Optional[UUID] = None
        comment: Optional[str] = None
        creator_uuid: Optional[UUID] = None
        filename: Optional[str] = None
        import_status: Optional[str] = None
        labels: Optional[List[str]] = field(default_factory=lambda: [])
        name: Optional[str] = None
        post_time: Optional[datetime] = None
        type: Optional[str] = None

        class Schema(MarshmallowSchema):
            id = fields.Integer(missing=None)
            uuid = fields.UUID(missing=None)
            comment = fields.String(missing=None)
            creator_uuid = fields.UUID(data_key="creatorUuid", missing=None)
            filename = fields.String(missing=None)
            import_status = fields.String(data_key="importStatus", missing=None)
            labels = fields.List(
                fields.String,
                many=True,
                missing=[],
                default=[],
                allow_none=True,
            )
            name = fields.String(missing=None)
            post_time = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ", data_key="postTime", missing=None
            )
            type = fields.String(missing=None)

            @post_load
            def make_source(self, data, **kwargs):
                return HiveLibrary.Source(**data)

    @dataclass
    class IP:
        id: Optional[int] = None
        ip: Optional[IPv4Address] = None
        uuid: Optional[UUID] = None

        class Schema(MarshmallowSchema):
            id = fields.Integer(load_only=True, missing=None, default=None)
            ip = fields.IPv4(load_only=True, missing=None, default=None)
            uuid = fields.UUID(load_only=True, missing=None, default=None)

            class Meta:
                unknown = EXCLUDE

            @post_load
            def make_ip(self, data, **kwargs):
                return HiveLibrary.IP(**data)

    @dataclass
    class File:
        parent_id: Optional[int] = None
        id: Optional[int] = None
        uuid: Optional[UUID] = None
        create_time: Optional[datetime] = None
        creator_uuid: Optional[UUID] = None
        node_id: Optional[int] = None
        caption: Optional[str] = None
        control_sum: Optional[str] = None
        name: Optional[str] = None
        size: Optional[int] = None
        mime_type: Optional[str] = None

        class Schema(MarshmallowSchema):
            parent_id = fields.Integer(
                load_only=True, missing=None, default=None, data_key="parentId"
            )
            id = fields.Integer(load_only=True, missing=None, default=None)
            uuid = fields.UUID(missing=None, default=None)
            create_time = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
                load_only=True,
                data_key="createTime",
                missing=None,
            )
            creator_uuid = fields.UUID(
                load_only=True, missing=None, default=None, data_key="creatorUuid"
            )
            node_id = fields.Integer(missing=None, default=None, data_key="nodeId")
            caption = fields.String(missing=None, default=None)
            control_sum = fields.String(
                missing=None, default=None, data_key="controlSum"
            )
            name = fields.String(missing=None, default=None, data_key="filename")
            size = fields.Integer(missing=None, default=None, data_key="filesize")
            mime_type = fields.String(missing=None, default=None, data_key="mimetype")

            class Meta:
                unknown = EXCLUDE

            @pre_load(pass_many=False)
            def pre_make_file(self, data, **kwargs):
                if "mimeType" in data:
                    data["mimetype"] = data["mimeType"]
                    del data["mimeType"]
                return data

            @post_load
            def make_file(self, data, **kwargs):
                return HiveLibrary.File(**data)

    @dataclass
    class Note:

        def __str__(self):
            return self.text

        parent_id: Optional[int] = None
        id: Optional[int] = None
        uuid: Optional[UUID] = None
        create_time: Optional[datetime] = None
        creator_uuid: Optional[UUID] = None
        text: Optional[str] = None
        files: Optional[List["HiveLibrary.File"]] = field(default_factory=lambda: [])
        notes: Optional[List["HiveLibrary.Note"]] = field(default_factory=lambda: [])
        tags: Optional[List["HiveLibrary.Tag"]] = field(default_factory=lambda: [])

        class Schema(MarshmallowSchema):
            parent_id = fields.Integer(
                load_only=True, missing=None, default=None, data_key="parentId"
            )
            id = fields.Integer(load_only=True, missing=None, default=None)
            uuid = fields.UUID(load_only=True, missing=None, default=None)
            create_time = fields.DateTime(
                "%Y-%m-%dT%H:%M:%S.%fZ",
                load_only=True,
                data_key="createTime",
                missing=None,
            )
            creator_uuid = fields.UUID(
                load_only=True, missing=None, default=None, data_key="creatorUuid"
            )
            text = fields.String(missing=None, default=None)
            files = fields.Nested(
                lambda: HiveLibrary.File.Schema,
                load_only=True,
                many=True,
                missing=[],
                default=[],
                allow_none=True,
            )
            notes = fields.Nested(
                lambda: HiveLibrary.Note.Schema,
                load_only=True,
                many=True,
                missing=[],
                default=[],
                allow_none=True,
            )
            tags = fields.Nested(
                lambda: HiveLibrary.Tag.Schema,
                load_only=True,
                many=True,
                missing=[],
                default=[],
            )

            class Meta:
                unknown = EXCLUDE

            @post_load
            def make_note(self, data, **kwargs):
                return HiveLibrary.Note(**data)

    @dataclass
    class Host:
        checkmarks: Optional[List["HiveLibrary.Checkmark"]] = field(
            default_factory=lambda: []
        )
        files: Optional[List["HiveLibrary.File"]] = field(default_factory=lambda: [])
        id: Optional[int] = None
        uuid: Optional[UUID] = None
        notes: Optional[List["HiveLibrary.Note"]] = field(default_factory=lambda: [])
        ip: Optional[IPv4Address] = None
        ip_binary: Optional[str] = None
        records: Optional[List["HiveLibrary.Record"]] = field(
            default_factory=lambda: []
        )
        names: Optional[List["HiveLibrary.Host.Name"]] = field(
            default_factory=lambda: []
        )
        ports: Optional[List["HiveLibrary.Host.Port"]] = field(
            default_factory=lambda: []
        )
        sources: Optional[List["HiveLibrary.Source"]] = field(
            default_factory=lambda: []
        )
        tags: Optional[List["HiveLibrary.Tag"]] = field(default_factory=lambda: [])

        class Schema(MarshmallowSchema):
            checkmarks = fields.Nested(
                lambda: HiveLibrary.Checkmark.Schema,
                load_only=True,
                many=True,
                data_key="checkmarks",
                missing=[],
                default=[],
            )
            files = fields.Nested(
                lambda: HiveLibrary.File.Schema,
                many=True,
                data_key="files",
                missing=[],
                default=[],
                allow_none=True,
            )
            id = fields.Integer(
                load_only=True, missing=None, default=None, data_key="id"
            )
            uuid = fields.UUID(
                load_only=True, missing=None, default=None, data_key="uuid"
            )
            notes = fields.Nested(
                lambda: HiveLibrary.Note.Schema,
                load_only=True,
                many=True,
                data_key="notes",
                missing=[],
                default=[],
                allow_none=True,
            )
            ip = fields.IPv4(missing=None, default=None, allow_none=True)
            ip_binary = fields.String(
                load_only=True, missing=None, default=None, data_key="ipBinary"
            )
            records = fields.Nested(
                lambda: HiveLibrary.Record.Schema,
                many=True,
                missing=[],
                default=[],
            )
            names = fields.Nested(
                lambda: HiveLibrary.Host.Name.Schema,
                many=True,
                data_key="hostnames",
                missing=[],
                default=[],
            )
            ports = fields.Nested(
                lambda: HiveLibrary.Host.Port.Schema,
                many=True,
                missing=[],
                default=[],
            )
            sources = fields.Nested(
                lambda: HiveLibrary.Source.Schema,
                many=True,
                missing=[],
                default=[],
            )
            tags = fields.Nested(
                lambda: HiveLibrary.Tag.Schema,
                many=True,
                data_key="tags",
                missing=[],
                default=[],
            )

            class Meta:
                unknown = EXCLUDE

            @post_dump(pass_many=False)
            def parse_host(self, data, many, **kwargs):
                if data["ip"] is not None:
                    data["ipv4"] = data["ip"]
                del data["ip"]
                return data

            @pre_load(pass_many=False)
            def pre_make_records(self, data, **kwargs):
                if "records" in data:
                    if isinstance(data["records"], Dict):
                        records: Dict[str, List] = data["records"]
                        new_records: List[Dict] = list()
                        for key, value_list in records.items():
                            for value in value_list:
                                new_records.append(value)
                        data["records"] = new_records
                if "ip" in data:
                    if data["ip"] == "":
                        data["ip"] = None
                return data

            @post_load
            def make_host(self, data, **kwargs):

                return HiveLibrary.Host(**data)

        @dataclass
        class Name:
            checkmarks: Optional[List["HiveLibrary.Checkmark"]] = field(
                default_factory=lambda: []
            )
            files: Optional[List["HiveLibrary.File"]] = field(
                default_factory=lambda: []
            )
            id: Optional[int] = None
            ips: Optional[List["HiveLibrary.IP"]] = None
            uuid: Optional[UUID] = None
            notes: Optional[List["HiveLibrary.Note"]] = field(
                default_factory=lambda: []
            )
            hostname: Optional[str] = None
            records: Optional[List["HiveLibrary.Record"]] = field(
                default_factory=lambda: []
            )
            sources: Optional[List["HiveLibrary.Source"]] = field(
                default_factory=lambda: []
            )
            tags: Optional[List["HiveLibrary.Tag"]] = field(default_factory=lambda: [])

            class Schema(MarshmallowSchema):
                checkmarks = fields.Nested(
                    lambda: HiveLibrary.Checkmark.Schema,
                    load_only=True,
                    many=True,
                    data_key="checkmarks",
                    missing=[],
                    default=[],
                )
                files = fields.Nested(
                    lambda: HiveLibrary.File.Schema,
                    many=True,
                    data_key="files",
                    missing=[],
                    default=[],
                    allow_none=True,
                )
                id = fields.Integer(
                    load_only=True, missing=None, default=None, data_key="id"
                )
                ips = fields.Nested(
                    lambda: HiveLibrary.IP.Schema,
                    many=True,
                    data_key="ips",
                    missing=[],
                    default=[],
                    allow_none=True,
                    load_only=True,
                )
                uuid = fields.UUID(
                    load_only=True, missing=None, default=None, data_key="uuid"
                )
                notes = fields.Nested(
                    lambda: HiveLibrary.Note.Schema,
                    load_only=True,
                    many=True,
                    data_key="notes",
                    missing=[],
                    default=[],
                    allow_none=True,
                )
                hostname = fields.String(data_key="hostname", missing=None)
                records = fields.Nested(
                    lambda: HiveLibrary.Record.Schema,
                    many=True,
                    data_key="records",
                    missing=[],
                    default=[],
                )
                sources = fields.Nested(
                    lambda: HiveLibrary.Source.Schema,
                    many=True,
                    missing=[],
                    default=[],
                )
                tags = fields.Nested(
                    lambda: HiveLibrary.Tag.Schema,
                    many=True,
                    data_key="tags",
                    missing=[],
                    default=[],
                    allow_none=True,
                )

                class Meta:
                    unknown = EXCLUDE

                @pre_load(pass_many=False)
                def pre_make_records(self, data, **kwargs):
                    if "records" in data:
                        if isinstance(data["records"], Dict):
                            records: Dict[str, List] = data["records"]
                            new_records: List[Dict] = list()
                            for key, value_list in records.items():
                                for value in value_list:
                                    new_records.append(value)
                            data["records"] = new_records
                    return data

                @post_load
                def make_hostname(self, data, **kwargs):
                    return HiveLibrary.Host.Name(**data)

        @dataclass
        class Port:
            ip: Optional[IPv4Address] = None
            checkmarks: Optional[List["HiveLibrary.Checkmark"]] = field(
                default_factory=lambda: []
            )
            files: Optional[List["HiveLibrary.File"]] = field(
                default_factory=lambda: []
            )
            id: Optional[int] = None
            uuid: Optional[UUID] = None
            notes: Optional[List["HiveLibrary.Note"]] = field(
                default_factory=lambda: []
            )
            port: Optional[int] = None
            service: Optional["HiveLibrary.Host.Port.Service"] = None
            protocol: str = "tcp"
            state: str = "open"
            records: Optional[List["HiveLibrary.Record"]] = field(
                default_factory=lambda: []
            )
            sources: Optional[List["HiveLibrary.Source"]] = field(
                default_factory=lambda: []
            )
            tags: Optional[List["HiveLibrary.Tag"]] = field(default_factory=lambda: [])

            class Schema(MarshmallowSchema):
                ip = fields.IPv4(missing=None, default=None)
                checkmarks = fields.Nested(
                    lambda: HiveLibrary.Checkmark.Schema,
                    load_only=True,
                    many=True,
                    missing=[],
                    default=[],
                )
                files = fields.Nested(
                    lambda: HiveLibrary.File.Schema,
                    many=True,
                    missing=[],
                    default=[],
                    allow_none=True,
                )
                id = fields.Integer(load_only=True, missing=None, default=None)
                uuid = fields.UUID(load_only=True, missing=None, default=None)
                notes = fields.Nested(
                    lambda: HiveLibrary.Note.Schema,
                    load_only=True,
                    many=True,
                    missing=[],
                    default=[],
                    allow_none=True,
                )
                port = fields.Integer(missing=None)
                service = fields.Nested(
                    lambda: HiveLibrary.Host.Port.Service.Schema,
                    missing=[],
                    default=[],
                )
                protocol = fields.String(missing=None, default="tcp")
                state = fields.String(missing=None, default="open")
                records = fields.Nested(
                    lambda: HiveLibrary.Record.Schema,
                    many=True,
                    missing=[],
                    default=[],
                )
                sources = fields.Nested(
                    lambda: HiveLibrary.Source.Schema,
                    many=True,
                    missing=[],
                    default=[],
                )
                tags = fields.Nested(
                    lambda: HiveLibrary.Tag.Schema,
                    many=True,
                    missing=[],
                    default=[],
                    allow_none=True,
                )

                class Meta:
                    unknown = EXCLUDE

                @pre_load(pass_many=False)
                def pre_make_records(self, data, **kwargs):
                    if "records" in data:
                        if isinstance(data["records"], Dict):
                            records: Dict[str, List] = data["records"]
                            new_records: List[Dict] = list()
                            for key, value_list in records.items():
                                for value in value_list:
                                    new_records.append(value)
                            data["records"] = new_records
                    return data

                @post_load
                def make_port(self, data, **kwargs):
                    return HiveLibrary.Host.Port(**data)

            @dataclass
            class Service:
                name: Optional[str] = None
                product: Optional[str] = None
                version: Optional[str] = None
                cpelist: Optional[str] = None

                class Schema(MarshmallowSchema):
                    name = fields.String(missing=None)
                    product = fields.String(missing=None)
                    version = fields.String(missing=None)
                    cpelist = fields.String(missing=None)

                    @post_dump(pass_many=False)
                    def clean_missing(self, data, many, **kwargs):
                        clean_data = data.copy()
                        for key in filter(lambda key: data[key] is None, data):
                            del clean_data[key]
                        return clean_data

                    @post_load
                    def make_service(self, data, **kwargs):
                        return HiveLibrary.Host.Port.Service(**data)

    @staticmethod
    def load_config() -> Config:
        config: HiveLibrary.Config = HiveLibrary.Config()
        if path.isfile(config.file) and path.getsize(config.file) > 0:
            with open(config.file, "r") as config_file:
                config_dict: Dict[str, str] = safe_load(config_file)
                config = HiveLibrary.Config.Schema().load(config_dict)
        return config

    @staticmethod
    def dump_config(config: Config) -> None:
        hive_config_dir: str = path.dirname(config.file)
        if not path.isdir(hive_config_dir):
            makedirs(hive_config_dir)
        with open(config.file, "w") as config_file:
            config_dict: Dict[str, str] = HiveLibrary.Config.Schema().dump(config)
            safe_dump(config_dict, config_file)



# Description
"""
Hive REST API console client
Author: HexWay
License: MIT
Copyright 2021, HexWay
"""

# Import
from hive_library import HiveLibrary
from hive_library.rest import HiveRestApi, AuthenticationError
from hive_library.enum import RecordTypes
from argparse import ArgumentParser
from time import sleep
from uuid import UUID
from ipaddress import IPv4Address
from typing import Optional, List
from datetime import datetime
from prettytable import PrettyTable

# Authorship information
__author__ = "HexWay"
__copyright__ = "Copyright 2021, HexWay"
__credits__ = [""]
__license__ = "MIT"
__version__ = "0.0.1b1"
__maintainer__ = "HexWay"
__email__ = "contact@hexway.io"
__status__ = "Development"


# Pretty print hosts
def pretty_print_hosts(hosts: List[HiveLibrary.Host]) -> None:
    keys: List[str] = [
        "Id",
        "IP",
        "Hostname",
        "Tags",
        "Port id",
        "Port",
        "Proto",
        "Service",
        "Product",
        "Port tags",
    ]
    table: PrettyTable = PrettyTable(keys)
    for host in hosts:
        for port in host.ports:
            table.add_row(
                [
                    host.id,
                    host.ip,
                    ", ".join(str(name.hostname) for name in host.names),
                    ", ".join(str(tag.name) for tag in host.tags),
                    port.id,
                    port.port,
                    port.protocol,
                    port.service.name,
                    port.service.product + " " + port.service.version,
                    ", ".join(str(tag.name) for tag in port.tags),
                ]
            )
    print(table)


# Pretty print projects
def pretty_print_projects(projects: List[HiveLibrary.Project]) -> None:
    keys: List[str] = ["Id", "Name", "Description", "Start Date", "End Date"]
    table: PrettyTable = PrettyTable(keys)
    for project in projects:
        table.add_row(
            [
                project.id,
                project.name,
                project.description,
                project.start_date,
                project.end_date,
            ]
        )
    print(table)


# Main function
def main() -> None:
    # region Parse script arguments
    parser: ArgumentParser = ArgumentParser(description="Hive REST API console client")

    # Hive arguments
    parser.add_argument(
        "-S", "--server_url", type=str, help="set Hive server URL", default=None
    )
    parser.add_argument(
        "-U", "--username", type=str, help="set Hive username", default=None
    )
    parser.add_argument(
        "-P", "--password", type=str, help="set Hive password", default=None
    )
    parser.add_argument(
        "-I", "--project_id", type=UUID, help="set project UUID", default=None
    )

    # Create project
    parser.add_argument(
        "--create_project", action="store_true", help="create Hive project"
    )
    parser.add_argument(
        "--project_name", type=str, help="set created Hive project name", default=None
    )
    parser.add_argument(
        "--project_description",
        type=str,
        help="set created Hive project description",
        default=None,
    )
    parser.add_argument(
        "--project_group",
        type=UUID,
        help="set created Hive project group UUID",
        default=None,
    )
    parser.add_argument(
        "--project_start",
        type=str,
        help="set created Hive project start date, example: 2021-01-01 01:01:01",
        default=None,
    )
    parser.add_argument(
        "--project_end",
        type=str,
        help="set created Hive project end date, example: 2021-02-01 01:01:01",
        default=None,
    )

    # Create host
    parser.add_argument("--create_host", action="store_true", help="create Hive host")
    parser.add_argument(
        "--host_tag", type=str, help="set created Hive host tag", default=None
    )
    parser.add_argument(
        "--host_name", type=str, help="set created Hive host name", default=None
    )
    parser.add_argument(
        "--host_ip",
        type=IPv4Address,
        help="set created Hive host IP address",
        default=None,
    )
    parser.add_argument(
        "--host_record_name",
        type=str,
        help="set created Hive host record name",
        default=None,
    )
    parser.add_argument(
        "--host_record_tool",
        type=str,
        help="set created Hive host record tool name",
        default=None,
    )
    parser.add_argument(
        "--host_record_val",
        type=str,
        help="set created Hive host record value",
        default=None,
    )
    parser.add_argument(
        "--port",
        type=int,
        help="set port for created Hive host",
        default=None,
    )
    parser.add_argument(
        "--port_tag", type=str, help="set port tag for created Hive host", default=None
    )
    parser.add_argument(
        "--port_proto",
        type=str,
        help="set port protocol for created Hive host",
        default="tcp",
    )
    parser.add_argument(
        "--port_state",
        type=str,
        help="set port state for created Hive host",
        default="open",
    )
    parser.add_argument(
        "--port_service_cpelist",
        type=str,
        help="set service cpelist in port for created Hive host",
        default=None,
    )
    parser.add_argument(
        "--port_service_name",
        type=str,
        help="set service name in port for created Hive host",
        default=None,
    )
    parser.add_argument(
        "--port_service_product",
        type=str,
        help="set service product in port for created Hive host",
        default=None,
    )
    parser.add_argument(
        "--port_service_version",
        type=str,
        help="set service version in port for created Hive host",
        default=None,
    )
    parser.add_argument(
        "--port_record_name",
        type=str,
        help="set record name for port in created Hive host",
        default=None,
    )
    parser.add_argument(
        "--port_record_tool",
        type=str,
        help="set record tool name for port in created Hive host",
        default=None,
    )
    parser.add_argument(
        "--port_record_val",
        type=str,
        help="set record value for port in created Hive host",
        default=None,
    )

    # Get objects
    parser.add_argument(
        "--get_projects", action="store_true", help="get Hive projects list"
    )
    parser.add_argument(
        "--get_groups", action="store_true", help="get Hive groups list"
    )
    parser.add_argument(
        "--get_hosts", action="store_true", help="get Hive hosts list for project"
    )
    parser.add_argument(
        "--get_host", type=int, help="get Hive host by identifier", default=None
    )
    parser.add_argument(
        "--get_port", type=int, help="get Hive port by identifier", default=None
    )

    # Search
    parser.add_argument(
        "--search_ip",
        type=IPv4Address,
        help="Search hosts by IP address",
        default=None,
    )
    parser.add_argument(
        "--search_port",
        type=int,
        help="Search hosts by port",
        default=None,
    )
    parser.add_argument(
        "--search_tag",
        type=str,
        help="Search hosts by tag",
        default=None,
    )
    parser.add_argument(
        "--search_service",
        type=str,
        help="Search hosts by service",
        default=None,
    )

    # Delete
    parser.add_argument(
        "--delete_project", type=str, help="delete Hive project by name", default=None
    )

    # Proxy
    parser.add_argument("-p", "--proxy", type=str, help="Set proxy URL", default=None)
    args = parser.parse_args()
    # endregion

    # Init HiveRestApi
    try:
        hive_api: HiveRestApi = HiveRestApi(
            username=args.username,
            password=args.password,
            server=args.server_url,
            proxy=args.proxy,
            project_id=args.project_id,
        )
    except AuthenticationError as error:
        print(f"Authentication Error: {error}")
        exit(1)

    # Get projects
    if args.get_projects:
        projects: Optional[List[HiveLibrary.Project]] = hive_api.get_projects_list()
        if len(projects) == 0:
            print("Not found projects!")
        else:
            print("Projects:")
            pretty_print_projects(projects=projects)
        exit(0)

    # Get groups
    if args.get_groups:
        groups: Optional[List[HiveLibrary.Group]] = hive_api.get_groups()
        if groups is None:
            print("Not found groups!")
        else:
            keys: List[str] = ["Id", "Name", "Description", "Projects"]
            table: PrettyTable = PrettyTable(keys)
            for group in groups:
                table.add_row(
                    [
                        group.id,
                        group.name,
                        group.description,
                        ", ".join(str(project.name) for project in group.projects),
                    ]
                )
            print(f"Groups:\n{table}")
        exit(0)

    # Create project
    if args.create_project:
        new_project: HiveLibrary.Project = HiveLibrary.Project()
        if args.project_name is not None:
            new_project.name = args.project_name
        else:
            print(
                "Hive project name is not set! "
                "Please set Hive project name argument: --project_name"
            )
            exit(4)
        if args.project_description is not None:
            new_project.description = args.project_description
        if args.project_group is not None:
            new_project.group_id = args.project_group
        if args.project_start is not None:
            new_project.start_date = datetime.strptime(args.project_start, "%Y-%m-%d")
        if args.project_end is not None:
            new_project.end_date = datetime.strptime(args.project_end, "%Y-%m-%d")
        created_project: Optional[HiveLibrary.Project] = hive_api.create_project(
            project=new_project
        )
        if created_project is None:
            print(f"Failed to create new project:")
        else:
            print(f"Successfully created new project:")
        keys: List[str] = ["Id", "Name", "Description", "Start Date", "End Date"]
        table: PrettyTable = PrettyTable(keys)
        table.add_row(
            [
                new_project.id,
                new_project.name,
                new_project.description,
                new_project.start_date,
                new_project.end_date,
            ]
        )
        print(table)
        exit(0)

    # Check project id is not None
    config: HiveLibrary.Config = HiveLibrary.load_config()
    if args.project_id is None:
        project_id = config.project_id
    else:
        project_id = args.project_id

    if project_id is None:
        print(
            "Hive project id is not set! "
            "Please set Hive project id in config file or argument: --project_id"
        )
        exit(3)

    # Create host
    if args.create_host:
        new_host: HiveLibrary.Host = HiveLibrary.Host()
        if args.host_ip is not None:
            new_host.ip = args.host_ip
        if args.host_name is not None:
            new_host.names = [HiveLibrary.Host.Name(hostname=args.host_name)]
        if args.host_tag is not None:
            new_host.tags = [HiveLibrary.Tag(name=args.host_tag)]
        if (
            args.host_record_name is not None
            and args.host_record_tool is not None
            and args.host_record_val is not None
        ):
            new_host.records = [
                HiveLibrary.Record(
                    name=args.host_record_name,
                    tool_name=args.host_record_tool,
                    record_type=RecordTypes.STRING.value,
                    value=args.host_record_val,
                )
            ]
        if args.port is not None:
            new_host.ports = [
                HiveLibrary.Host.Port(
                    port=args.port, protocol=args.port_proto, state=args.port_state
                )
            ]
        if args.port_tag is not None:
            new_host.ports[0].tags = [HiveLibrary.Tag(name=args.port_tag)]
        if (
            args.port_record_name is not None
            and args.port_record_tool is not None
            and args.port_record_val is not None
        ):
            new_host.ports[0].records = [
                HiveLibrary.Record(
                    name=args.port_record_name,
                    tool_name=args.port_record_tool,
                    record_type=RecordTypes.STRING.value,
                    value=args.port_record_val,
                )
            ]
        new_host.ports[0].service = HiveLibrary.Host.Port.Service()
        if args.port_service_cpelist is not None:
            new_host.ports[0].service.cpelist = args.port_service_cpelist
        if args.port_service_name is not None:
            new_host.ports[0].service.name = args.port_service_name
        if args.port_service_product is not None:
            new_host.ports[0].service.product = args.port_service_product
        if args.port_service_version is not None:
            new_host.ports[0].service.version = args.port_service_version
        task_id: Optional[UUID] = hive_api.create_host(
            project_id=project_id, host=new_host
        )
        if task_id is None:
            print(f"Failed to create new host:")
        else:
            print("Creating new host ...")
            for _ in range(30):
                if hive_api.task_is_completed(project_id=project_id, task_id=task_id):
                    break
                else:
                    sleep(1)
            if hive_api.task_is_completed(project_id=project_id, task_id=task_id):
                print(f"Successfully created new host:")
            else:
                print(f"Failed to create new host:")
        keys: List[str] = [
            "IP",
            "Hostname",
            "Port",
            "Proto",
            "Service",
            "Product",
            "Tags",
        ]
        table: PrettyTable = PrettyTable(keys)
        table.add_row(
            [
                new_host.ip,
                ", ".join(str(name.hostname) for name in new_host.names),
                new_host.ports[0].port,
                new_host.ports[0].protocol,
                new_host.ports[0].service.name,
                new_host.ports[0].service.product
                + " "
                + new_host.ports[0].service.version,
                ", ".join(str(tag.name) for tag in new_host.tags),
            ]
        )
        print(table)

    # Get hosts
    if args.get_hosts:
        hosts: Optional[List[HiveLibrary.Host]] = hive_api.get_hosts(
            project_id=project_id
        )
        if hosts is None or len(hosts) == 0:
            print(f"Not found hosts in project: {project_id}")
        else:
            print(f"Hosts in project: {project_id}")
            pretty_print_hosts(hosts=hosts)

    # Get host by id
    if args.get_host is not None:
        host: Optional[HiveLibrary.Host] = hive_api.get_host(
            project_id=project_id, host_id=int(args.get_host)
        )
        if host is None:
            print(f"Not found host with id: {args.get_host} in project: {project_id}")
        else:
            print(f"Host id: {args.get_host} in project: {project_id}")
            keys: List[str] = [
                "Id",
                "IP",
                "Hostname",
                "Tags",
                "Record id",
                "Record tool",
                "Record name",
                "Record value",
            ]
            table: PrettyTable = PrettyTable(keys)
            for record in host.records:
                if isinstance(record.value, List) and isinstance(
                    record.value[0], HiveLibrary.Record
                ):
                    for child_record in record.value:
                        table.add_row(
                            [
                                host.id,
                                host.ip,
                                ", ".join(str(name.hostname) for name in host.names),
                                ", ".join(str(tag.name) for tag in host.tags),
                                child_record.id,
                                child_record.import_type,
                                child_record.name,
                                child_record.value,
                            ]
                        )
                else:
                    table.add_row(
                        [
                            host.id,
                            host.ip,
                            ", ".join(str(name.hostname) for name in host.names),
                            ", ".join(str(tag.name) for tag in host.tags),
                            record.id,
                            record.import_type,
                            record.name,
                            record.value,
                        ]
                    )
            print(table)

    # Search by IP address
    if args.search_ip is not None:
        hosts: Optional[List[HiveLibrary.Host]] = hive_api.search_by_ip(
            project_id=project_id, ip=args.search_ip
        )
        if hosts is None or len(hosts) == 0:
            print(
                f"Not found hosts in project: {project_id} with ip address: {args.search_ip}"
            )
        else:
            print(f"Hosts in project: {project_id} with ip address: {args.search_ip}")
            pretty_print_hosts(hosts=hosts)

    # Search by port
    if args.search_port is not None:
        hosts: Optional[List[HiveLibrary.Host]] = hive_api.search_by_port(
            project_id=project_id, port=args.search_port
        )
        if hosts is None or len(hosts) == 0:
            print(
                f"Not found hosts in project: {project_id} with port: {args.search_port}"
            )
        else:
            print(f"Hosts in project: {project_id} with port: {args.search_port}")
            pretty_print_hosts(hosts=hosts)

    # Search by tag
    if args.search_tag is not None:
        hosts: Optional[List[HiveLibrary.Host]] = hive_api.search_by_tag(
            project_id=project_id, tag=args.search_tag
        )
        if hosts is None or len(hosts) == 0:
            print(
                f"Not found hosts in project: {project_id} with tag: {args.search_tag}"
            )
        else:
            print(f"Hosts in project: {project_id} with tag: {args.search_tag}")
            pretty_print_hosts(hosts=hosts)

    # Search by service
    if args.search_service is not None:
        hosts: Optional[List[HiveLibrary.Host]] = hive_api.search_by_service(
            project_id=project_id, service=args.search_service
        )
        if hosts is None or len(hosts) == 0:
            print(
                f"Not found hosts in project: {project_id} with service: {args.search_service}"
            )
        else:
            print(f"Hosts in project: {project_id} with service: {args.search_service}")
            pretty_print_hosts(hosts=hosts)

    # Delete project
    if args.delete_project is not None:
        deleted_project: Optional[
            HiveLibrary.Project
        ] = hive_api.delete_project_by_name(project_name=args.delete_project)
        if deleted_project is None:
            print(f"Not found project with name: {args.delete_project}")
        else:
            print("Deleted project:")
            pretty_print_projects(projects=[deleted_project])


# Call Main function
if __name__ == "__main__":
    main()



# Description
"""
Hive enums
Author: HexWay
License: MIT
Copyright 2021, HexWay
"""

# Import
from enum import Enum


# Authorship information
__author__ = "HexWay"
__copyright__ = "Copyright 2021, HexWay"
__credits__ = [""]
__license__ = "MIT"
__version__ = "0.0.1b1"
__maintainer__ = "HexWay"
__email__ = "contact@hexway.io"
__status__ = "Development"


class ListEnum(Enum):
    @classmethod
    def list(cls):
        return list(map(lambda c: c.value, cls))


class PermissionTypes(str, ListEnum):
    EDIT = "EDIT"
    READONLY = "READONLY"
    OWNER = "OWNER"
    ADMIN = "ADMIN"


class RecordTypes(str, ListEnum):
    STRING = "string"
    TEXT_BLOCK = "text_block"
    NUMBER = "number"
    TUPLE = "tuple"
    CODEBLOCK = "codeblock"
    FILE = "file"
    LIST = "list"
    DIRECTORY = "directory"
    NESTED = "nested"


class TaskStates(str, ListEnum):
    SUCCESS = "SUCCESS"
    PROGRESS = "PROGRESS"
    PENDING = "PENDING"
    FAILURE = "FAILED"
    CANCELLED = "CANCELLED"



# Description
"""
setup.py: Python Hive Library setup
Author: HexWay
License: MIT
Copyright 2021, HexWay
"""

# Import
from setuptools import setup, find_packages

# Authorship information
__author__ = "HexWay"
__copyright__ = "Copyright 2021, HexWay"
__credits__ = [""]
__license__ = "MIT"
__version__ = "0.0.1b1"
__maintainer__ = "HexWay"
__email__ = "contact@hexway.io"
__status__ = "Development"

# Setup
with open("README.md", "r") as readme:
    long_description = readme.read()

setup(
    name="hive_library",
    version=__version__,
    author=__author__,
    author_email=__email__,
    description="Python Hive Library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://hexway.io/",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Education",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Education",
        "Topic :: Security",
    ],
    install_requires=[
        "requests",
        "urllib3",
        "dataclasses",
        "marshmallow",
        "PyYAML",
        "prettytable",
    ],
    entry_points={
        "console_scripts": ["hive-cli=hive_library.cli:main"],
    },
    python_requires=">=3.6",
    include_package_data=True,
)
